﻿using C1.Win.C1FlexGrid;
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Runtime.Remoting.Metadata;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;

namespace FlexAutoSizeRow
{
  public partial class Form1 : Form
  {
    public Form1()
    {
      InitializeComponent();

      //"UseCompatibleTextRendering" = false might cause clipped texts.
      //Console.WriteLine(this.c1FlexGrid1.UseCompatibleTextRendering);
      this.c1FlexGrid1.UseCompatibleTextRendering = true;

      this.c1FlexGrid1.DrawMode = DrawModeEnum.OwnerDraw;
      this.c1FlexGrid1.OwnerDrawCell += C1FlexGrid1_OwnerDrawCell;

      this.c1FlexGrid1.Cols[3].Width = 295;

      //Build a long string:
      string todisplay = "";
      const int MAX = 100;
      for (int index = 0; index < MAX; index++)
      {
        if (index > 0)
        {
          todisplay += " // ";
        }
        //with underscores, the word gets longer and thus causes trouble with splitting. More problems if "UseCompatibleTextRendering" = false.
        //todisplay += "part " + index + " of " + MAX;
        todisplay += "part_" + index + "_of_" + MAX;
      }
      
      this.c1FlexGrid1[2, 0] = "row 2";
      this.c1FlexGrid1[2, 3] = todisplay;

      CellStyle styleWrap = this.c1FlexGrid1.Styles.Add("Wrap", this.c1FlexGrid1.Styles.Normal);
      styleWrap.WordWrap = true;
      this.c1FlexGrid1.SetCellStyle(2, 3, styleWrap);

      this.c1FlexGrid1[3, 0] = "row 3";
      //Text has to be set to ALL cells of the merged range.
      this.c1FlexGrid1[3, 3] = todisplay;
      this.c1FlexGrid1.SetCellStyle(3, 3, styleWrap);

      this.c1FlexGrid1[4, 0] = "row 4";
      this.c1FlexGrid1[4, 3] = todisplay;
      this.c1FlexGrid1.SetCellStyle(4, 3, styleWrap);

      this.c1FlexGrid1[5, 0] = "row 5";
      this.c1FlexGrid1[5, 3] = todisplay;
      this.c1FlexGrid1.SetCellStyle(5, 3, styleWrap);

      this.c1FlexGrid1.AllowMerging = AllowMergingEnum.Custom;
      //First the "shorter" range:
      this.c1FlexGrid1.MergedRanges.Add(this.c1FlexGrid1.GetCellRange(2, 3, 5, 3));

      //some merged text in col 1:
      string shortmerged = "This is a quite short text which should merge and wrap!";
      this.c1FlexGrid1[2, 2] = shortmerged;
      this.c1FlexGrid1.SetCellStyle(2, 2, styleWrap);
      this.c1FlexGrid1[3, 2] = shortmerged;
      this.c1FlexGrid1.SetCellStyle(3, 2, styleWrap);
      this.c1FlexGrid1.MergedRanges.Add(this.c1FlexGrid1.GetCellRange(2, 2, 3, 2));

      //some merged text in col 0:
      shortmerged = "This is a quite short text which should merge and wrap!";
      this.c1FlexGrid1[3, 1] = shortmerged;
      this.c1FlexGrid1.SetCellStyle(3,1, styleWrap);
      this.c1FlexGrid1.MergedRanges.Add(this.c1FlexGrid1.GetCellRange(3, 1, 4, 1));

      //Unmerged text in row:
      string shortunmerged = "This is a quite short text which should wrap!";
      this.c1FlexGrid1[2, 1] = shortunmerged;
      this.c1FlexGrid1.SetCellStyle(2,1, styleWrap);
    }

    private void C1FlexGrid1_OwnerDrawCell(object sender, OwnerDrawCellEventArgs e)
    {
      if (e.Row == 2 && e.Col == 3 && e.Measuring == true)
      {
        //Console.WriteLine("Measuring!");
        //e.DrawCell();

        SizeF size = e.Graphics.MeasureString(e.Text, this.c1FlexGrid1.Font, this.c1FlexGrid1.Cols[3].WidthDisplay);
        Console.WriteLine("Measured size: " + size);
      }
    }

    private void buttonWorkaround_Click(object sender, EventArgs e)
    {
      // Autosize row, ignoring merged cells.
      int bottomRow = c1FlexGrid1.Rows.Count - 1;
      int rightCol = c1FlexGrid1.Cols.Count - 1;
      var flag = AutoSizeFlags.IgnoreHidden | AutoSizeFlags.IgnoreMerged;
      c1FlexGrid1.AutoSizeRows(0, 0, bottomRow, rightCol, 0, flag);
            
      // Get all merged ranges.
      var ranges = c1FlexGrid1.MergedRanges;

      foreach(CellRange range in ranges)
      {
        // Store old sizes (this will be its minimum sizes).
        Row lastRow = null;
        // Use Dictionary to increase performance (use c1FlexGrid1.Rows[n] only once).
        var rows = new Dictionary<Row, int>();
        for (int i = range.r1; i <= range.r2; i++)
        {
          // Find and save last row.
          lastRow = c1FlexGrid1.Rows[i];
          rows.Add(lastRow, lastRow.HeightDisplay);
        }

        // Autosize range taking into account merged cells.
        c1FlexGrid1.AutoSizeRows(range.r1, range.c1, range.r2, range.c2, 0, AutoSizeFlags.None);
                
        // Delta size for the last row.
        int delta = 0;
        foreach (var row in rows)
        {
          // Looking for rows that have changed the size, but are not the last row.
          if (row.Key != lastRow && row.Value != row.Key.Height)
          {
            // Store delta size of the row.
            delta += row.Key.HeightDisplay - row.Value;
            // Return back the old size for all merged rows except the last one.
            row.Key.Height = row.Value;
          }
        }

        // Now we can set size for the last row using delta.
        // And it should not be smaller than its original size.
        lastRow.Height = Math.Max(rows[lastRow], lastRow.HeightDisplay + delta);
      }

      //this.c1FlexGrid1.AutoSizeRows();
      //this.c1FlexGrid1.AutoSizeRows(2, 3, 4, 3, 0, AutoSizeFlags.None);
      //this.c1FlexGrid1.AutoSizeRows(2, 3, 4, 3, 0, AutoSizeFlags.SameSize);

    }

    private void buttonHeightToLast_Click(object sender, EventArgs e)
    {
      //First do default Autosize - so that I can merge autosized rows:
      this.c1FlexGrid1.AutoSizeRows();

      //Store height of each merged range: 
      List<KeyValuePair<CellRange, int>> listRange2Height = new List<KeyValuePair<CellRange, int>>();
      foreach (CellRange rangeMerged in this.c1FlexGrid1.MergedRanges)
      {
        //Only if range is larger than one row!
        if (rangeMerged.BottomRow != rangeMerged.TopRow)
        {
          //Unfortunartely, AutoSizeRows for this range.
          this.c1FlexGrid1.AutoSizeRows(rangeMerged.TopRow, rangeMerged.LeftCol, rangeMerged.BottomRow, rangeMerged.RightCol, 0, AutoSizeFlags.None);
          
          //Store heigth of all rows:
          int sumRowHeight = 0;
          for (int row = rangeMerged.TopRow; row <= rangeMerged.BottomRow; row++)
          {
            sumRowHeight += this.c1FlexGrid1.Rows[row].HeightDisplay;
          }

          listRange2Height.Add(new KeyValuePair<CellRange, int>(rangeMerged, sumRowHeight));
        }

      }
      
      //TODO: Sort ranges from top to bottom / left to right?!

      //Now autoSize all rows AGAIN - without merged ranges:
      this.c1FlexGrid1.AutoSizeRows(0, 0, this.c1FlexGrid1.Rows.Count, this.c1FlexGrid1.Cols.Count, 0, AutoSizeFlags.IgnoreMerged);

      //For each merge range: set the the height of the last row to "remaining heigth of merged range".
      foreach (KeyValuePair<CellRange, int> range2SumHeight in listRange2Height)
      {
        int sumHeight = range2SumHeight.Value;

        CellRange range = range2SumHeight.Key;
        //Remove HeightDisplay of all other rows (but not the height of the bottom row itself!)
        for (int rowInRange = range.TopRow; rowInRange < range.BottomRow; rowInRange++)
        {
          sumHeight -= this.c1FlexGrid1.Rows[rowInRange].HeightDisplay;
        }

        //Remaining height:
        //Apply only if the height is bigger then the default height (by AutoSizeRow)
        if (sumHeight > this.c1FlexGrid1.Rows[range.BottomRow].HeightDisplay)
        {
          this.c1FlexGrid1.Rows[range.BottomRow].Height = sumHeight;
        }
      }
    }

    private void buttonReset_Click(object sender, EventArgs e)
    {
      for (int row = 0; row < this.c1FlexGrid1.Rows.Count; row++)
      {
        this.c1FlexGrid1.Rows[row].Height = -1;
      }
    }
  }
}
